from .base import Messager

__ignore__ = True


class NakonakoMessager(Messager):
    name = "NakoNako"
    chat_name = "NakoNetwork"
    default_messages = [
        "watery-wl@v1.yaml * 10",
        "goodday-wl@v1.yaml",
        "goodnoon-wl@v1.yaml",
        "goodeve-wl@v1.yaml",
        "goodnight-wl@v1.yaml",
    ]
